<div{% include 'django/forms/widgets/attrs.html' %}>
  <p>
    {% for entry in summary %}
    <strong>{{ entry.label }}</strong>{% if entry.value %}: <bdi>{{ entry.value }}</bdi>{% endif %}
    {% endfor %}
  </p>
  <p><a class="button" href="{{ password_url|default:"../password/" }}">{{ button_label }}</a></p>
</div>
