body {
    background: url("#foobar");
    background: url("http:foobar");
    background: url("https:foobar");
    background: url("data:foobar");
    background: url("chrome:foobar");
    background: url("//foobar");
    background: url();
}

